// command line to compile:
// sdcc -mz180 --data-loc0xC000 --code-loc0x8000 --no-std-crt0 4megFlashProgrammer.c  
// This is a program for a Z8S180 to program, verify, and read data 
// with a 4 meg FLASH memory chip at 0x0000 & ram at 0x8000
// stack 0xFF0F set in crt0.rel
// runs at 115200 baud and uses an Intel Hex file to load program
// sent to Z8S180 board with ascii file transfer.
//   to begin ascii file transfer in minicom: cntrl a, s, select "ascii" from
//   pulldown menu, use down arrow to find file, select with spacebar, press enter


#include <z8s180.h>

volatile char *address;  // address pointer to a byte in memory
volatile int addr;       // address variable that can be shifted to construct a address
volatile char rflag = 0; // serial character received indicator flag
volatile unsigned char ch = 0; // the serial character received
      // global strings are stored more efficiently by sdcc than local strings  
const char msgE[]       = {"\n\r Erase FLASH"};
const char endmsg[]     = {" Complete"}; 
const char startmsg[]   = {"\n\r Enter starting address of program - press Enter to run\n\r "};
const char runmsg[]     = {" - Running Program\n\r"};
const char rangemsg[]   = {"\n\r Programming Failed ! - An address is above 0x7FFF !"};
const char begintrans[] = {"\n\r Begin ASCII transfer of Intel hex file - press Esc when done "};
const char complete[]   = {"\n\r Programming complete\n\r"}; 
const char failmsg[]    = {"\n\r Verify Failed"}; 
const char passmsg[]    = {"\n\r Verify Passed"};
const char entermsg[]   = {"\n\r Enter start address - Press Enter to begin reading - Esc to stop\n\r"};
const char cmpltmsg[]   = {"\n\r\n\r RAM read complete"};  
const char menu[]       = {"\n\r Press: 'p' to Program FLASH, 'v' to Verify Program in FLASH\n\r        'r' to Read Memory,   'G' to run Program\n\r        'E' to erase FLASH\n\r"};
const char enablemsg[]  = {"\n\r Write Enable FLASH before programming"};
const char contmsg[]    = {"\n\r Press any key to continue - Esc to end "};
      // function prototypes
void serial_interrupt_handler(void) __interrupt 0x0B;
void PrintS(char *str);
void PrintC(char pchar);
void program_verify(char flag);
void read_translate(); 
void readram();
void run_program();
void enter_addr();
void Flash_erase();

     // main is first function so that program will begin at code-loc0x8000
void main()
{  
  char status0;
  int *vectorpointer; // pointer to interrupt vector table ( at 0xFFc0 ) 

    __asm              
    LD SP, #0xFFD0    // set stack pointer and initialize some registers
    LD HL, #0x0000
    LD DE, #0X0000
    LD BC, #0X0000
    LD IX, #0X0000 
    LD IY, #0X0000
    __endasm;    
    
                  // Setup interrupt vector for ASCI0    
    vectorpointer = ((int *)0xFFE0 + ASCI0_VECTOR); // add ASCI0(0x0E) offset to pointer address 
    *vectorpointer = (int)serial_interrupt_handler; // save address of ISR at 0xFFE0
                      // 00    20    40    60    80    A0    C0    E0 = possible IL values
    IL = 0xE0;        // set interrupt vector table lower address to 0x20 - 001? ????             
                      // 000   001  010   011   100   101   110   111
    __asm             // the interrupt Mode, Enable, and I reg need to be set with assembly
    IM 2              // set interrupt mode 2          
    LD A, #0xFF       // set interrupt vector table upper address byte to 0xFF
    LD I, A           // load 0xFF into I register
    EI                // enable interrupts  
    __endasm;   
     
    rflag = 0;  
                  // set up CPU and Clock 
    OMCR    = 0;      // OPERATION MODE CONTROL REGISTER IS SET TO 0
    CCR     = 0X80;   // CLOCK DIVIDED BY 0 gives 33,000,000 Mhz phi
    DCNTL  |= 0x80;   // 2 wait states for 70ns ROM
    RCR     = 0;      // 0 dissables refresh FOR STATIC RAM: reset value = 1
                  // set up ASCIO      
    CNTLA0  = 0X74;   // 0111 0100  RE(6) = 1, TE(5) = 1, rts0(4) = 1, 8n1(2) = 1 
    status0 = STAT0;  // read status control register 05H - read DCD0 pin to enable receive
    ASEXT0  = 0x68;   // 0110 1000 - disable DCD0(6) and CTS0(5); X1(4) = 0; BRG0(3) Mode = 1
    CNTLB0  = 0x00;   // 0000 0000 - prescale 10 PS(5) = 0, DR(3) = 0,    
    STAT0  |= 0x08;   // 0000 1000 enable receive interrupt rei(3) = 1         
                  // set up baud rate                    phi = 33,000,000 Hz
                      // 0x1AC = 428 =   2400 baud: TC = phi / (2 x   2400 x 16) -2  
                      // 0xD5  = 213 =   4800 baud: TC = phi / (2 x   4800 x 16) -2     
                      // 0x69  = 105 =   9600 baud: TC = phi / (2 x   9600 x 16) -2 
    ASTC0L  = 0x07;   // 0x34  =  52 =  19200 baud: TC = phi / (2 x  19200 x 16) -2
    ASTC0H  = 0x00;   // 0x10  =  16 =  57600 baud: TC = phi / (2 x  57600 x 16) -2
                      // 0x07  =   7 = 115200 baud: TC = phi / (2 x 115200 x 16) -2          
                      
    PrintS(enablemsg);
    PrintS(menu);

    while(1)
      { 
      while(rflag == 0); // wait for a keypress
        
      rflag = 0;         // key pressed! so zero receive flag
      
      if(ch == 'p')
         program_verify('p');
      if(ch == 'v')
         program_verify('v');
      if(ch == 'r')        
         readram();
      if(ch == 'G') 
         run_program(); 
      if(ch == 'E') 
         Flash_erase();
      if(ch == 0x1b || ch == 0x0d)
         PrintS(menu);          
      }  
}

            // INTERUPT SERVICE ROUTINE for interrupt 11(0x0b) ASCI0
void serial_interrupt_handler(void) __interrupt 0x0B 
{ 
    CNTLA0 &= 0xF7;   // 1111 0111 clear error flag(EFR-3) in CNTLA0
    ch = RDR0;        // get serial input (iy used to save ch to RAM)
    rflag = 1;        // (iy used to save rflag to RAM)
    CNTLA0 = CNTLA0 ^ 0X10; // toggle RTS(4) LED
    
    STAT0 |= 0x08;    // 0000 1000 enable receive interrupt
    __asm
    EI                // enable interrupts
    __endasm;
} // end ISR -----------------------------------------------------------


void readram()   // prints contents of ram in hex format
{
  char lines = 0;
  unsigned int lines_read = 0;  
    
    PrintS(entermsg);
    enter_addr();

    if(ch != 0x1b) 
       address = (char *)addr;
    else
       PrintS(menu);

    while(ch != 0x1b) 
      { 
      read_translate();
      if(lines++ == 31)
         {
         PrintS(contmsg);
         while(rflag == 0); // wait for a keypress
         lines = 0;
         rflag = 0;         // key pressed ! zero receive flag        
         } 
      if(lines_read++ > 0x7fe)
         break;
      }
    PrintS(cmpltmsg);
    PrintS(menu); 
    ch = 0; 
}


void read_translate()  // convert binary in memory to ASCII serial port output
{ 
  int convert;
  int row = 0; 
  char pchar = 0;
    
      PrintS("\n\r");  // start a new line to translate
                      // print the starting address of the line
      convert = (int)address;     // 1111 1111 0000 0000
      convert = (convert >> 8);   // ff00 >> 8 00ff
      convert = (convert & 0x00ff);
      pchar   = (char)convert;
      PrintC(pchar);  // print 2 left digits of address
        
      convert = (int)address;
      convert = (convert & 0X00ff);
      pchar   = (char)convert;
      PrintC(pchar);  // print 2 right digits of address
   
       
      while(( STAT0 & 0x02 ) != 0x02); // wait for transmitter empty
      
      TDR0 = 0x20;     // send space after address
     
      while( row < 16) 
        {
        pchar = (unsigned char)*address;
        PrintC(pchar);
    
        while(( STAT0 & 0x02 ) != 0x02); // wait for transmitter empty
        
        TDR0 = 0x20;   // send space after 2 digits
       
        address++;     // go to next character
        row++;         // count characters ( 16 per row )
        
        if(rflag == 1) // if a key is pressed during screen dump read it
          {
          rflag = 0;
          if(ch == 0x1b) // if it is Esc - stop the screen dump
             break; 
          }
        }     
} 


void PrintC(char pchar)  // prints pchar in hex format
{
  unsigned char ldigit;
  unsigned char rdigit;
    
    rdigit = (pchar & 0x000f);   

    pchar = (pchar & 0x00f0);
    ldigit = (pchar >> 4); // xy >> 4 0x
    
    if((ldigit & 0x0f) < 10)
      ldigit += 0x30;
    else
      ldigit += 0x37;
    
    while(( STAT0 & 0x02 ) != 0x02); // wait for transmitter empty
    
    TDR0 = ldigit;     
     
    if((rdigit & 0x0f) < 10)
      rdigit += 0x30;
    else
      rdigit += 0x37;
    
    while(( STAT0 & 0x02 ) != 0x02); // wait for transmitter empty
    
    TDR0 = rdigit;     
         
    CNTLA0 = CNTLA0 ^ 0X10; // toggle RTS(4) LED 
}


void PrintS(char *str)       // prints text string 
{ 
    while(*str != 0)
      {
      if(( STAT0 & 0x02 ) == 0x02)
         {
         TDR0 = *str++;   
         CNTLA0 = CNTLA0 ^ 0X10; // toggle RTS(4) LED
         }
      }
}
 

void program_verify(char flag)  // 2643 clock cycles available between charactes
{ 
  volatile int addr = 0;
  char test = 0;
  char hex = 0;
  char position = 0;
  char count = 0; 
  char type = 1;
     
    PrintS(begintrans);
    
    ch = 0;    // zero ch since it is tested before character is received
    while(ch != 0x1b)
      {
      if(rflag == 1)           // rflag set by receive interrupt
        {
        rflag = 0;             // clear the interrupt receive flag
        if(ch == 0x1b)
           break;

        position++;            // go to next position of ihx format
       
        if(ch == ':')          // start of ihx line identifier
           position = 0;       // set position to position 0 = ihx line length(left)
        if(ch > 0x40)          // A is 0x41, a is 0x61  
           ch = ch + 9;        // add 9 to convert ch from ascii to hexadecimal                       
        ch = (0x0f & ch);      // strip the left digit
      
        if(position == 1)      // ihx line length(left)
           count = (ch << 4);  // count = the number code bytes in the ihx line 
        if(position == 2)      // ihx line length(right)     
           count = (ch | count);
     
        if(position == 3)      // decode starting address for the ihx line of code
           {    
           addr = (int)ch;
           addr = addr << 4;   // most significant digit of address
           }
        if(position == 4)
           {
           addr |= (int)ch;
           addr = addr << 4;
           }
        if(position == 5)
           {
           addr |= (int)ch;
           addr = addr << 4;
           }
        if(position == 6)
           {
           addr |= (int)ch;   // least significant digit of address 
           address = (char *)addr;  // address complete
           }

        if(position == 7)     // decode type of line
           { 
           type =1; 
           type = (ch << 4);  // count is converted to use 
           }
        if(position == 8)     // to identify data positions      
           type = (ch | type);

        if(position > 8)      // convert ascii code to hex code     
           if(type == 0)    
              if( (position & 0x01) == 0x01) // check even/odd position
                 hex = (ch << 4);  // odd = left nibble bits 4-7
              else                 // even = low nibble bits 0-3   
                 if((count--) > 0) // if this is still code in the ihx line 
                    {
                    if(flag == 'p') 
                       if(address < (char *)0x8000)
                         {
                         __asm        // UNPROTECT BYTE SEQUENCE

                         LD A, #0xAA
                         LD (#0x5555), A
 			 
                         LD A, #0x55
                         LD (#0x2AAA), A
 			 
                         LD A, #0xA0
                         LD (#0x5555), A
           
                         __endasm;

                         hex = (hex | ch); // convert the 2 ascii digits to hex value 
                         *address = hex;
                         address++;    // increment address to next address in ihx line
                         }
                       else
                         {
                         flag = 'o';
                         ch = 0x1b;
                         } 
                    if(flag == 'v')
                       {
                       hex = (hex | ch);
                       test = *address;
                       address++;
                       if(test != hex)         
                          flag = 'f';
                       }
                    }
        }  // end if rflag == 1 
      }    // end while ch != 0x1b  
    if(flag == 'o')
       PrintS(rangemsg);  
    if(flag == 'p')
       PrintS(complete);
    if(flag == 'f')
       PrintS(failmsg);
    if(flag == 'v')
       PrintS(passmsg);
    PrintS(menu); 
}


void run_program()
{   
    PrintS(startmsg);
    enter_addr();
  
    if(ch != 0x1b)
       {  
       PrintS(runmsg);             
       __asm
       LD HL, (#_addr)    // put start of program address in HL
       JP (HL)            // indirect jump to address in HL 
       __endasm;         
       }
}


void enter_addr()
{
  char digits = 4;

    while(ch != 0x0d && ch != 0x1b)  // if keypress 5 == 'g' end loop
      {
      rflag = 0;
      if(digits == 4) // reset "digits" and start entry loop again
         digits = 0;
     
      while(digits < 4)
         { 
         while(rflag == 0); // wait for keypress 1
      
         rflag = 0;
         if(ch == 0x1b)
           break;
         if(( STAT0 & 0x02 ) == 0x02) // echo keypress 1
           TDR0 = ch;
         if(ch > 0x40)      // A is 0x41, a is 0x61  
           ch = ch + 9;     // add 9 to convert ch from ascii to hexadecimal                       
         addr = (addr << 4);
         addr |= (int)(0x000f & ch);  
         digits++;   // = 1
         } 
      
      if(ch != 0x1b)
         while(rflag == 0); // wait for keypress 5
       
      rflag = 0;     
      }           // end while ch != 'g'
}


void Flash_erase()
{   
    PrintS(msgE);

    rflag = 0;
    while(rflag == 0);
    rflag = 0;
    
    if(ch != 0x1b)
       {
       __asm
       LD A, #0XAA
       LD (#0x5555), A  
    
       LD A, #0X55
       LD (#0X2AAA), A

       LD A, #0X80
       LD (#0x5555), A 
   
       LD A, #0XAA 
       LD (#0x5555), A
    
       LD A, #0X55
       LD (#0X2AAA), A

       LD A, #0X10
       LD (#0X5555), A 
       __endasm;
   
       PrintS(endmsg);
       } 
   PrintS(menu);  
}
