// command line to compile:
// sdcc -mz180 --data-loc0xFFD0 4megrom.c
// FFD0 to FFFF gives 47 bytes of RAM - 10 bytes used 
// 0xFFB0 = Stack pointer set in crt0.rel
// 0xFFC0 = Interrupt Vector Table - 16 bytes - ASCI0 vector @ FFCE-FFCF
// 0XFFD0 = Ram data location - 47 bytes
// This is a program for a Z8S180 to program, verify, and read data 
// with a X28C256 32k EEPROM memory chip at starting address 0x0000 
// and a KM62256 32K static RAM at 0x8000
// 
// The program is compiled with sdcc, with 2 custom modifications.
//    crt0.rel file was modified with stack location above.
//    z180.h was modified to add names and addresses of additional 
//    sfr's that the Z8S180 CPU has over the z180 sfr's    
// 
// The program connects with a terminal that runs at 57600 baud
// It uploads Intel Hex program files using ascii serial file transfer
//    To begin ascii file transfer in minicom: cntrl a, s, select "ascii" from
//    pulldown menu, use down arrow to find file, select with spacebar, press enter
//
// Command line to compile code for RAM:
// sdcc -mz180 --data-loc0xF000 --code-loc0x8000 --no-std-crt0 ramprog.c
// RAM (80ns) program can run with 1 wait state DCNTL = 0x70;
// EEPROM (150ns or 250ns) needs 3 wait states - default on power up (DCNTL = 0xF0) 

#include <z8s180.h>

volatile char *address;  
volatile char rflag = 0; 
volatile unsigned char ch = 0;
volatile int addr;
volatile char hex;
volatile char test = 0;
volatile char position = 0;
volatile char count = 0; 
volatile char type = 1;
volatile char flag;

const char entermsg[] = {"\n\r Enter start address - Press Enter to read - Esc to stop\n\r"};
const char cmpltmsg[] = {"\n\r\n\rRead Complete\n\r"};
const char changebaud[] = {"\n\rChange baud rate to 2400 - press any key to continue\n\r"};
const char restorebaud[] = {"\n\rChange baud back to 57600 - press any key to continue\n\r"};
const char enablemsg[] = {"\n\r Write Enable EEPROM before programming\n\r"};
const char begintransfer[] = {"\n\rBegin ASCII transfer of hex file - press Esc when done\n\r"};
const char complete[] = {"\n\rProgramming complete\n\r"}; 
const char failmsg[] = {"\n\rVerify Failed"}; 
const char passmsg[] = {"\n\rVerify Passed"}; 
const char rangemsg[] = {"\n\rProgramming was blocked ! - An address was below 0x0B00 !\n\r"};
const char addressmsg[] = {"\n\rEnter starting address of program - press Enter to run\n\r  "};
const char runmsg[] = {"\n\rRunning Program\n\r"};
const char contmsg[] = {"\n\rPress any key to continue - Esc to end "};
const char menu[] = {"\n\r Press:\n\r 'p' To Program RAM\n\r 'v' To Verify Program\n\r 'r' To Read Memory\n\r 'G' To run Program\n\r"};                 //'E' To Program EEPROM\n\r"};
// unsigned char routine[] = {0x2A,0xD0,0XFF,0X3A,0XD4,0XFF,0x77,0X3A,0XD6,0XFF,0XB7,0X28,0XFA,0XC9}; 

void serial_interrupt_handler(void) __interrupt 0x0B;
void PrintS(char *str);
void PrintC(char pchar);
char program_verify();
void read_translate(); 
void run_program();
void enter_addr();
void program_eeprom();
void program();
void verify();
void setup_mem();


void main()
{  
  char status0;
  int *vectorpointer; // pointer to interrupt vector table ( at 0xFFc0 ) 

    
    OMCR    = 0;      // OPERATION MODE CONTROL REGISTER IS SET TO 0
    RCR     = 0;      // 0 dissables refresh FOR STATIC RAM: reset value = 1
    CCR     = 0x80;   // CLOCK DIVIDED BY 0 = 0x80 divided by 2 = 0x00 
                      // clock set to 33 Mhz
   // DCNTL  |= 0x80; // 2 memory wait states (0xF0 default)
                      // Setup interrupt vector for ASCI0
    vectorpointer = ((int *)0xFFC0 + ASCI0_VECTOR); // add ASCI0(0x0E) offset to pointer address 
    *vectorpointer = (int)serial_interrupt_handler; // save address of ISR at 0xFFCE
                      // 00    20    40    60    80    A0    C0    E0 = possible IL values
    IL = 0xC0;        // set interrupt vector table lower address to 0xC0 - 110? ???? 
                      // 000   001  010   011   100   101   110   111
    __asm             
    IM 2              // set interrupt mode 2          
    LD A, #0xFF       // set interrupt vector table upper address byte to 0xFF
    LD I, A           // load 0xFB into I register
    EI                // enable interrupts  
    __endasm;         
         
    CNTLA0  = 0X74;   // 0111 0100  RE(6) = 1, TE(5) = 1, rts0(4) = 1, 8n1(2) = 1 
    status0 = STAT0;  // read status control register 05H - read DCD0 pin to enable receive
    ASEXT0  = 0x68;   // 0110 1000 - disable DCD0(6) and CTS0(5); X1(4) = 0; BRG0(3) Mode = 1
    CNTLB0  = 0x00;   // 0000 0000 - prescale 10 PS(5) = 0, DR(3) = 0, 
                                                  //     phi = 33,000,000
    ASTC0L  = 0x07;   // 0x1AC = 428 =   2400 baud: TC = phi / (2 x   2400 x 16) -2 
    ASTC0H  = 0x00;   // 0xD5  = 213 =   4800 baud: TC = phi / (2 x   4800 x 16) -2
                      // 0x69  = 105 =   9600 baud: TC = phi / (2 x   9600 x 16) -2 
                      // 0x34  =  52 =  19200 baud: TC = phi / (2 x  19200 x 16) -2 
                      // 0x10  =  16 =  57600 baud: TC = phi / (2 x  57600 x 16) -2 
                      // 0x07  =   7 = 115200 baud: TC = phi / (2 x 115200 x 16) -2          
   
    STAT0  |= 0x08;   // 0000 1000 enable receive interrupt rei(3) = 1
    
    PrintS(menu);
    while(1)          // wait for a menu command character
      { 
      while(rflag == 0); // wait for a keypress
        
      rflag = 0;      // key was pressed, reset the receive flag
      
      if(ch == 'p')
         program();
     // if(ch == 'E')
     //    program_eeprom();
      if(ch == 'v')
         verify();
      if(ch == 'r')        
         read_translate();
      if(ch == 'G')
         run_program(); 
      if(ch == 0x1b || ch == 0x0d)
         PrintS(menu);       
      }  
}

 // INTERUPT SERVICE ROUTINE for interrupt 11(0x0B) ASCI0 -------------

void serial_interrupt_handler(void) __interrupt 0x0B 
{ 
      CNTLA0 &= 0xF7;   // 1111 0111 clear error flag(EFR-3) in CNTLA0
      ch = RDR0;        // get serial input 
      rflag = 1;        // rflag to indicate ch received
    
      CNTLA0 = CNTLA0 ^ 0X10; // toggle RTS(4) LED
    
      STAT0 |= 0x08;    // 0000 1000 enable receive interrupt
      __asm
      EI                // enable interrupts
      __endasm;
}
 // end ISR -----------------------------------------------------------


void read_translate()  // convert binary in memory to ASCII serial port output
{ 
  char lines = 0;
  int row; 

      PrintS(entermsg);
      enter_addr();    // enter the address to start reading memory

      if(ch != 0x1b)   // exits if Esc pressed during enter_addr()
         {
         address = (char *)addr;
         while(ch != 0x1b)
            {              
            PrintS("\n\r"); // start a new line to translate
                          // print the starting address of the line
            PrintC((int)address >> 8); // print 2 lefthand characters  of address         
            PrintC((int)address); // print 2 righthand characters of address   
           
            for(row = 0; row < 16; row++, address ++) // print 1 row of 16 bytes
               {
               while(( STAT0 & 0x02 ) != 0x02); // wait for transmitter empty       
                 TDR0 = 0x20;    // send space before 2 digits
               PrintC(*address); // print 2 data digits at address
               }     //  end of a printing a row
            if(lines++ == 31)
               {
               PrintS(contmsg);
               while(rflag == 0); // wait for a keypress
               lines = 0;
               rflag = 0;         // key pressed ! zero receive flag        
               } 
            if(address == (char *)0x0000) // end read = top of ram 
               ch = 0x1b;    // this will trigger exit at next while loop cycle
            }     // while ch != 0x1b
         }     // end if ch != 0x1b
      PrintS(cmpltmsg);
      rflag = 0;    // rflag = 0 keeps from printimg menu twice
}       


void PrintC(char pchar) // prints hex pchar as 2 ascii characters
{
  unsigned char digit;
  char(l_r);

    digit = ((pchar & 0x00f0) >> 4); // set up left digit
    
    for(l_r = 0; l_r < 2; l_r++) // print left then right character 
      {
      digit += 0x30;    // add 0x30 to convert to ascii numeral
      if(digit > 0x39)  // if character is a letter  
        digit += 0x07;  // add 0x37 to convert to ascii letter    
     
      while(( STAT0 & 0x02 ) != 0x02); // wait for transmitter empty
      TDR0 = digit;     // then send the character to the terminal
     
      digit = (pchar & 0x000f);  // set up right digit
      }  
}


void PrintS(char *str)  // prints a string of text 
{ 
    while(*str != 0)
      if(( STAT0 & 0x02 ) == 0x02)
        TDR0 = *str++;   
}
 


void program()
{
    PrintS(begintransfer);
    
    flag = 'p';
    position = 0;
    ch = 0;            // zero ch since it is tested before character is received
// ----------------------------------------------
    while(ch != 0x1b)
      { 
      while(rflag == 0); // wait for received character
      
      rflag = 0;         // clear the interrupt receive flag
      if(ch == 0x1b)
          break;
      if(program_verify() == 1)    
          *address++ = hex; // save hex to address, increment address
      } 

    if(flag == 'o')
       PrintS(rangemsg);  
    if(flag == 'p')
       PrintS(complete);

    rflag = 0;
}


void verify()
{

 PrintS(begintransfer);
    
    flag = 'v';
    position = 0;
    ch = 0;    // zero ch since it is tested before character is received
    
    while(ch != 0x1b)
      { 
      while(rflag == 0);  // wait for received character
      
      rflag = 0;          // clear the interrupt receive flag
      if(ch == 0x1b)
         break;
      program_verify();    
      }
    
    if(flag == 'f')
       PrintS(failmsg);
    if(flag == 'v')
       PrintS(passmsg);

    rflag = 0;
}


char program_verify()
{ 
  char ret = 0;
  
      position++;         // go to next position of ihx format
       
      if(ch == ':')       // start of ihx line identifier
         position = 0;    // set position to position 0 = ihx line length(left)
      if(ch > 0x40)       // A is 0x41, a is 0x61  
         ch = ch + 9;     // add 9 to convert ch from ascii to hexadecimal                       
      ch = (0x0f & ch);   // strip the left digit
      
      if(position == 1)   // ihx line length(left)
         count = (ch << 4);  // count = the number code bytes in the ihx line 
      if(position == 2)   // ihx line length(right)     
         count = (ch | count);
     
      if(position == 3)   // decode starting address for the ihx line of cod   
         addr = (int)ch;
          
      if(position > 3 && position < 7) // positions 4,5,6
         {
         addr = addr << 4;
         addr |= (int)ch;
         if(position == 6) 
            address = (char *)addr;  // address complete
         }
         
      if(position == 7)      // decode type of line (left digit)
         type = (ch << 4);   // (00,01,02,03,04,05) are possibile types
           
      if(position == 8)      // decode type of line (right digit)     
         type = (ch | type); // 00 is a data byte

      if(position > 8)       // convert ascii code to hex code     
         if(type == 0)    
            if( (position & 0x01) == 0x01) // check even/odd position
               hex = (ch << 4);  // odd = left nibble bits 4-7
            else                 // even = low nibble bits 0-3   
               if((count--) > 0) // if this is still code in the ihx line 
                  {
                  if(flag == 'v')    // verify byte from ihx file
                     if(*address++ != (hex | ch))
                        flag = 'f';
                  if(flag == 'p')  // program byte from ihx file
                     if(address < (char *)0x0B00)  // protect the boot ROM program
                        flag = 'o';  // end - address conflicts with boot area                          
                     else
                        {
                        ret = 1;
                        hex = (hex | ch); // convert the 2 ascii digits to hex value                                       
                        }
                  }     
    return(ret);
}


void run_program()
{       
      PrintS(addressmsg);

      enter_addr();

      if(ch != 0x1b)
         {  
         PrintS(runmsg); 
            
         __asm
         LD HL, (#_addr)   // jump to address in _addr
         JP (HL)
         __endasm;
         }
}


void enter_addr()
{
  char digits = 4;

      while(ch != 0x0d && ch != 0x1b)  // if keypress 5 == 'g' end loop
        {
        rflag = 0;
        if(digits == 4) // reset "digits" and start entry loop again
           digits = 0;
     
        while(digits < 4)
           { 
           while(rflag == 0); // wait for keypress 1
      
           rflag = 0;
           if(ch == 0x1b)
              break;
           if(( STAT0 & 0x02 ) == 0x02) // echo keypress 1
              TDR0 = ch;
           if(ch > 0x40)      // A is 0x41, a is 0x61  
              ch = ch + 9;     // add 9 to convert ch from ascii to hexadecimal                       
           addr = (addr << 4);
           addr |= (int)(0x000f & ch);  
           digits++;   // = 1
           } 
      
        if(ch != 0x1b)
           while(rflag == 0); // wait for keypress 5
       
        rflag = 0;     
        }           // end while ch != 'g'
}


